/* notesldap.c -- check a password via LDAP to a notes server */

#include <ldap.h>
#include <stdio.h> /* printf */
#include <sys/time.h> /* struct timeval */
#include <stdlib.h> /* free */

char *ldapfinddn (const char *server, const char *searchdn, 
	const char *username) 
/* Find the LDAP DN for a given ID on the Lotus Domino servers */
{
  LDAP *ld;
  int r, errcode;
  char **attrs, *matcheddn, *errmsg, **referrals, /* *attr,*/ *dn;
  LDAPMessage *mess, *each;
  LDAPControl **serverctrls;
  struct timeval timeout;
  char *resultdn = NULL;
  /*BerElement *berp; */
  char un[200];

  ld = ldap_init (server,LDAP_PORT);
  if (ld==NULL) return NULL;

  r = ldap_simple_bind_s (ld,"","");
  if (r != LDAP_SUCCESS) {
    ldap_perror (ld,"bind failed");
    return NULL;
  }
  attrs = (char**) malloc (sizeof(char*) * 2);
  attrs[1]=NULL;
  attrs[0]="uid";
  timeout.tv_sec = 5;
  timeout.tv_usec = 0;
  snprintf (un,199,"uid=%s",username);
  un[199]=0;
  r = ldap_search_st (ld,searchdn,LDAP_SCOPE_SUBTREE,un,
	attrs,0,&timeout,&mess);
  free (attrs);
  if (r != LDAP_SUCCESS) {
    ldap_perror (ld,"search failed");
    return NULL;
  }
  /* r = ldap_count_messages (ld,mess);
  printf ("Returned %i messages\n",r); */
  each=ldap_first_message(ld,mess); 
  if (each) {
    r = ldap_parse_result (ld,each,&errcode,&matcheddn,&errmsg,&referrals,
	&serverctrls,0);  
    if (r || errcode) {
      if (errmsg) printf ("errmsg = %s\n",errmsg);
      if (r) ldap_perror (ld,"parse failed");
    } else {
      dn = ldap_get_dn (ld,each);
      if (dn && *dn) resultdn = strdup(dn);
      if (dn) ldap_memfree(dn);
      /* for (r=0; referrals && referrals[r]; r++) printf ("  ref(%i)=%s\n",
	r,referrals[r]);
      for (attr=ldap_first_attribute(ld,each,&berp); attr;
	attr=ldap_next_attribute (ld,each,berp)) {
        printf ("  attr=%s\n",attr);
      } */
    } 
    if (matcheddn) ldap_memfree(matcheddn);
    if (errmsg) ldap_memfree(errmsg);
    if (referrals) ldap_value_free(referrals);
    if (serverctrls) ldap_controls_free(serverctrls);
  } /* each exists */
  ldap_msgfree(mess);
  r = ldap_unbind(ld);
  if (r != LDAP_SUCCESS) {
    ldap_perror (ld,"unbind failed");
    if (resultdn) free (resultdn);
    return NULL;
  }
  ld = NULL;
  return resultdn;
}

char *ldappwbind (const char *server, const char *dn, const char *password) 
/* Find the LDAP DN for a given ID on the Lotus Domino servers */
{
  LDAP *ld;
  int r, errcode;
  char **attrs, *matcheddn, *errmsg, **referrals, *attr, **vals;
  LDAPMessage *mess, *each;
  LDAPControl **serverctrls;
  struct timeval timeout;
  BerElement *berp; 
  char *uid = NULL;

  ld = ldap_init (server,LDAP_PORT);
  if (ld==NULL) return NULL;

  r = ldap_simple_bind_s (ld,dn,password);
  if (r != LDAP_SUCCESS) { /* bad password */
    /* ldap_perror (ld,"bind failed"); */
    return NULL;
  }
  attrs = (char**) malloc (sizeof(char*) * 2);
  attrs[1]=NULL;
  attrs[0]="uid";
  timeout.tv_sec = 5;
  timeout.tv_usec = 0;
  r = ldap_search_st (ld,dn,LDAP_SCOPE_SUBTREE,"(objectclass=*)",
	attrs,0,&timeout,&mess);
  free (attrs);
  if (r != LDAP_SUCCESS) {
    ldap_perror (ld,"search failed");
    return NULL;
  }
  /* r = ldap_count_messages (ld,mess);
  printf ("Returned %i messages\n",r); */
  each=ldap_first_message(ld,mess); 
  if (each) {
    r = ldap_parse_result (ld,each,&errcode,&matcheddn,&errmsg,&referrals,
	&serverctrls,0);  
    if (r || errcode) {
      if (errmsg) printf ("errmsg = %s\n",errmsg);
      if (r) ldap_perror (ld,"parse failed");
    } else {
      /* for (r=0; referrals && referrals[r]; r++) printf ("  ref(%i)=%s\n",
	r,referrals[r]); */
      for (attr=ldap_first_attribute(ld,each,&berp); attr;
	attr=ldap_next_attribute (ld,each,berp)) {
        if (strcmp(attr,"uid")==0) {
	  vals = ldap_get_values(ld,each,attr);
          if (vals && vals[0]) {
	    uid = strdup (vals[0]);
	  }
	  if (vals) ldap_value_free(vals);
        }
        ldap_memfree(attr);
      } 
    } 
    if (matcheddn) ldap_memfree(matcheddn);
    if (errmsg) ldap_memfree(errmsg);
    if (referrals) ldap_value_free(referrals);
    if (serverctrls) ldap_controls_free(serverctrls);
  } /* each exists */
  ldap_msgfree(mess);
  r = ldap_unbind(ld);
  if (r != LDAP_SUCCESS) {
    ldap_perror (ld,"unbind failed");
    if (uid) free (uid);
    return NULL;
  }
  ld = NULL;
  return uid;
}



int main (int argc, char **argv) {
  char *dn, *uid;
 
  if (argc!=3) {
    fprintf (stderr,"Usage: %s <jbid> <NotesInternetPassword>\n",argv[0]);
    return 1;
  }
  dn = ldapfinddn("ldap.enterprise.server.com","O=BOC",argv[1]);
  if (dn && *argv[2]) {
    /* printf ("dn=%s\n",dn); */
    uid = ldappwbind ("ldap.enterprise.server.com",dn,argv[2]);
    if (!uid) return 1;
    /* else printf ("uid=%s\n",uid); */
  } else return 1;
  if (strcmp(uid,argv[1])!=0) return 1;
  return 0;
}
