/* notesldap.c -- check a password via LDAP to a notes server */

#include <ldap.h>
#include <stdio.h> /* printf */
#include <sys/time.h> /* struct timeval */
#include <stdlib.h> /* free */
#include <sys/types.h> /* getpwnam */
#include <pwd.h> /* getpwnam */

char *ldapfinddn (const char *server, const char *searchdn, 
	const char *username);
char *ldappwbind (const char *server, const char *dn, const char *password);

#include <security/_pam_macros.h>
/*
 * here, we make a definition for the externally accessible function
 * in this file (this definition is required for static a module
 * but strongly encouraged generally) it is used to instruct the
 * modules include file to define the function prototypes.
 */
#define PAM_SM_AUTH
#include <security/pam_modules.h>
/* --- authentication management functions (only) --- */

PAM_EXTERN
int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc,
                        const char **argv)
{
     int retval = PAM_SUCCESS;
     int minimum_uid = 500;
     char *pass=NULL;
     const char *name;
     int pamret;
     char *dn=NULL, *uid=NULL;

     if ((pamret = pam_get_user(pamh, &name, "login: ")) != PAM_SUCCESS) {
        return PAM_SERVICE_ERR;
     }
    if (minimum_uid>0) {
      /* Fail if minimum UID condition not met... */
      struct passwd     *pw = NULL;

      pw = getpwnam(name);
      if (!pw) {
        /* DLOG("getpwnam()", name); */
        return PAM_USER_UNKNOWN;
      }
      if (pw->pw_uid<minimum_uid) {
        /* DLOG("uid_too_low", name); */
        return PAM_SERVICE_ERR;
      }
    }

    (void) pam_get_item(pamh, PAM_AUTHTOK, (const void **) &pass);
    if (!pass) {
        /* DLOG("no_prior_password", name); */
	return PAM_AUTH_ERR;
    }

    dn = ldapfinddn("ldap.server.com","O=BOC",name);
    if (dn && *pass) {
      uid = ldappwbind ("ldap.server.com",dn,pass);
      if (!uid) retval = PAM_AUTH_ERR;
      else if (strcmp(uid,name)!=0) retval = PAM_SERVICE_ERR;
    } else retval = PAM_SERVICE_ERR;
    if (dn) free(dn);
    if (uid) free(uid);
    /* if (pass) free(pass); would double-free */

    return retval;
}

PAM_EXTERN
int pam_sm_setcred(pam_handle_t *pamh, int flags, int argc,
                   const char **argv)
{
     return PAM_SUCCESS;
}

#ifdef PAM_STATIC

/* static module data */

struct pam_module _pam_notesldap_modstruct = {
     "pam_notesldap",
     pam_sm_authenticate,
     pam_sm_setcred,
     NULL,
     NULL,
     NULL,
     NULL,
};

#endif




char *ldapfinddn (const char *server, const char *searchdn, 
	const char *username) 
/* Find the LDAP DN for a given username on the Lotus Domino servers */
{
  LDAP *ld;
  int r, errcode;
  char **attrs, *matcheddn, *errmsg, **referrals, /* *attr,*/ *dn;
  LDAPMessage *mess, *each;
  LDAPControl **serverctrls;
  struct timeval timeout;
  char *resultdn = NULL;
  /*BerElement *berp; */
  char un[200];

  ld = ldap_init (server,LDAP_PORT);
  if (ld==NULL) return NULL;

  r = ldap_simple_bind_s (ld,"","");
  if (r != LDAP_SUCCESS) {
    ldap_perror (ld,"bind failed");
    return NULL;
  }
  attrs = (char**) malloc (sizeof(char*) * 2);
  attrs[1]=NULL;
  attrs[0]="uid";
  timeout.tv_sec = 5;
  timeout.tv_usec = 0;
  snprintf (un,199,"uid=%s",username);
  un[199]=0;
  r = ldap_search_st (ld,searchdn,LDAP_SCOPE_SUBTREE,un,
	attrs,0,&timeout,&mess);
  free (attrs);
  if (r != LDAP_SUCCESS) {
    ldap_perror (ld,"search failed");
    return NULL;
  }
  /* r = ldap_count_messages (ld,mess);
  printf ("Returned %i messages\n",r); */
  each=ldap_first_message(ld,mess); 
  if (each) {
    r = ldap_parse_result (ld,each,&errcode,&matcheddn,&errmsg,&referrals,
	&serverctrls,0);  
    if (r || errcode) {
      if (errmsg) printf ("errmsg = %s\n",errmsg);
      if (r) ldap_perror (ld,"parse failed");
    } else {
      dn = ldap_get_dn (ld,each);
      if (dn && *dn) resultdn = strdup(dn);
      if (dn) ldap_memfree(dn);
      /* for (r=0; referrals && referrals[r]; r++) printf ("  ref(%i)=%s\n",
	r,referrals[r]);
      for (attr=ldap_first_attribute(ld,each,&berp); attr;
	attr=ldap_next_attribute (ld,each,berp)) {
        printf ("  attr=%s\n",attr);
      } */
    } 
    if (matcheddn) ldap_memfree(matcheddn);
    if (errmsg) ldap_memfree(errmsg);
    if (referrals) ldap_value_free(referrals);
    if (serverctrls) ldap_controls_free(serverctrls);
  } /* each exists */
  ldap_msgfree(mess);
  r = ldap_unbind(ld);
  if (r != LDAP_SUCCESS) {
    ldap_perror (ld,"unbind failed");
    if (resultdn) free (resultdn);
    return NULL;
  }
  ld = NULL;
  return resultdn;
}

char *ldappwbind (const char *server, const char *dn, const char *password) 
/* Find the LDAP DN for a given James Bond ID on the Lotus Domino servers */
{
  LDAP *ld;
  int r, errcode;
  char **attrs, *matcheddn, *errmsg, **referrals, *attr, **vals;
  LDAPMessage *mess, *each;
  LDAPControl **serverctrls;
  struct timeval timeout;
  BerElement *berp; 
  char *uid = NULL;

  ld = ldap_init (server,LDAP_PORT);
  if (ld==NULL) return NULL;

  r = ldap_simple_bind_s (ld,dn,password);
  if (r != LDAP_SUCCESS) { /* bad password */
    /* ldap_perror (ld,"bind failed"); */
    return NULL;
  }
  attrs = (char**) malloc (sizeof(char*) * 2);
  attrs[1]=NULL;
  attrs[0]="uid";
  timeout.tv_sec = 5;
  timeout.tv_usec = 0;
  r = ldap_search_st (ld,dn,LDAP_SCOPE_SUBTREE,"(objectclass=*)",
	attrs,0,&timeout,&mess);
  free (attrs);
  if (r != LDAP_SUCCESS) {
    ldap_perror (ld,"search failed");
    return NULL;
  }
  /* r = ldap_count_messages (ld,mess);
  printf ("Returned %i messages\n",r); */
  each=ldap_first_message(ld,mess); 
  if (each) {
    r = ldap_parse_result (ld,each,&errcode,&matcheddn,&errmsg,&referrals,
	&serverctrls,0);  
    if (r || errcode) {
      if (errmsg) printf ("errmsg = %s\n",errmsg);
      if (r) ldap_perror (ld,"parse failed");
    } else {
      /* for (r=0; referrals && referrals[r]; r++) printf ("  ref(%i)=%s\n",
	r,referrals[r]); */
      for (attr=ldap_first_attribute(ld,each,&berp); attr;
	attr=ldap_next_attribute (ld,each,berp)) {
        if (strcmp(attr,"uid")==0) {
	  vals = ldap_get_values(ld,each,attr);
          if (vals && vals[0]) {
	    uid = strdup (vals[0]);
	  }
	  if (vals) ldap_value_free(vals);
        }
        ldap_memfree(attr);
      } 
    } 
    if (matcheddn) ldap_memfree(matcheddn);
    if (errmsg) ldap_memfree(errmsg);
    if (referrals) ldap_value_free(referrals);
    if (serverctrls) ldap_controls_free(serverctrls);
  } /* each exists */
  ldap_msgfree(mess);
  r = ldap_unbind(ld);
  if (r != LDAP_SUCCESS) {
    ldap_perror (ld,"unbind failed");
    if (uid) free (uid);
    return NULL;
  }
  ld = NULL;
  return uid;
}



